@@ -1,8 +1,8 @@
 # *****************************************************************************
-# Â© Copyright IBM Corp. 2020.  All Rights Reserved.
+# © Copyright IBM Corp. 2020.  All Rights Reserved.
 #
 # This program and the accompanying materials
-# are made available under the terms of the Apache V2.0
+# are made available under the terms of the Apache V2.0 license
 # which accompanies this distribution, and is available at
 # http://www.apache.org/licenses/LICENSE-2.0
 #
@@ -426,7 +426,7 @@ class FileModelStore:
 
         if os.path.exists(filename):
             f = open(filename, "rb")
-            model = f.read(model)
+            model = f.read()
             f.close()
 
         if model is not None:
